function [ data, g, data0 ] = doubleIntegratorImplicit()
% doubleIntegratorImplicit: double integrator reach set (regular 2D HJ PDE)
%
%   [ data, g, data0 ] = doubleIntegratorImplicit()
%  
% This is a standard reachability calculation for the double integrator
% (albeit with nonstandard variable definition because we want to agree with
% the variable definitions for the partially nonlinear reachability paper):
%
%     \dot x = u   (velocity; the vertical dimension)
%     \dot y = x   (position; the horizontal dimension)
%        with input |u| \leq U and U is some constant.
%
% There are many options in the function which can only be modifed by
% changing the file.
%
% Input Parameters: none.
%
% Output Parameters:
%
%   data: Implicit surface function at t_max.
%
%   g: Grid structure on which data was computed.
%
%   data0: Implicit surface function at t_0.

% Copyright 2009 Ian M. Mitchell (mitchell@cs.ubc.ca).
% This software is used, copied and distributed under the licensing 
%   agreement contained in the file LICENSE in the top directory of 
%   the distribution.
%
% Created by Ian Mitchell, 9/01/09
% Subversion tags for version control purposes.
% $Date: 2011-04-11 10:47:05 -0700 (Mon, 11 Apr 2011) $
% $Id: doubleIntegratorImplicit.m 581 2011-04-11 17:47:05Z mitchell $

%---------------------------------------------------------------------------
% You will see many executable lines that are commented out.
%   These are included to show some of the options available; modify
%   the commenting to modify the behavior.
  
%---------------------------------------------------------------------------
%% Make sure we can see the kernel m-files.
run('addPathToKernel');

%---------------------------------------------------------------------------
%% PDE solver parameters.

% Integration parameters.
tMax = 4.0;                  % End time.
plotSteps = 9;               % How many intermediate plots to produce?
t0 = 0;                      % Start time.
singleStep = 0;              % Plot at each timestep (overrides tPlot).

% Period at which intermediate plots should be produced.
tPlot = (tMax - t0) / (plotSteps - 1);

% How close (relative) do we need to get to tMax to be considered finished?
small = 100 * eps;

% Pause after each plot?
pauseAfterPlot = 0;

% Delete previous plot before showing next?
deleteLastPlot = 0;

% Plot in separate subplots (set deleteLastPlot = 0 in this case)?
useSubplots = 0;

% Parameters for visualizeLevelSets.
level = 0;
displayType = 'contour';

%---------------------------------------------------------------------------
%% Optional and problem specific parameters.

% Maximum magnitude of the input.
input_bound = 0.25;

% What kind of target set?
target_set_type = 'rectangle'; %#ok<NASGU>
target_set_type = 'hexagon';

% Compute with both positive and negative velocities, or only positive?
negative_velocities = 1;

% Should we only allow the reach set to grow?
restrict_update = 1;

% How to implement the dynamics in the HJ PDE?
dynamics_term = 'convection';
% Could also implement as Lax-Freidrichs...

% Spatial approximation scheme.
derivFunc = @upwindFirstENO2;

% Temporal approximation scheme.
integratorFunc = @odeCFL2;

% Some integration parameters apply to any integrator.
integratorOptions = odeCFLset('factorCFL', 0.9, 'stats', 'on');
if(singleStep)
  integratorOptions = odeCFLset(integratorOptions, 'singleStep', 'on');
end

%---------------------------------------------------------------------------
%% Create the grid.  
g.dim = 2;
if negative_velocities
  g.min = 1.5 * [ -1; -1 ];
else
  g.min = 1.5 * [ -1;  0 ]; %#ok<UNRCH>
end
g.max = 1.5 * [ +1; +1 ];
g.N = 101;
g.bdry = @addGhostExtrapolate;
g = processGrid(g);

%---------------------------------------------------------------------------
%% Create initial conditions.
switch(target_set_type)

  case 'hexagon'
    % For this test, the target set is anything OUTSIDE of a hexagonal shape.
    % We use a hexagon instead of a square so that the maximum (and minimum)
    % position (y variable) for the target set is a continuous function of the 
    % velocity (x variable).  This continuity is not required for this Implicit
    % formulation, but will be important for the MIE formulation.

    % First, construct the hexagon.  Start with a rectangle to create the two
    % vertical sides.  If you just want a rectangular target set, you need not
    % include the extra intersections.
    data = shapeRectangleByCenter(g, 0, [ 2; 3 ]);
    data = shapeIntersection(data, shapeHyperplane(g, [ +1; +1 ], [ 0; +1.5 ]));
    data = shapeIntersection(data, shapeHyperplane(g, [ -1; +1 ], [ 0; +1.5 ]));
    data = shapeIntersection(data, shapeHyperplane(g, [ +1; -1 ], [ 0; -1.5 ]));
    data = shapeIntersection(data, shapeHyperplane(g, [ -1; -1 ], [ 0; -1.5 ]));
    
  case 'rectangle'
    % The rectangular target set cannot be done using the MIE formulation,
    % but it works fine for this Implicit formulation.
    data = shapeRectangleByCorners(g, [ -1.0; -1.0 ], [ +1.0, +1.0 ]);

  otherwise
    error('Unknown target_set_type: %s', target_set_type);
end
    
% Then take the complement, since the outside of the set is the actual
% target for the backwards reachability.
data = -data;
% If necessary, provide the IC as an output.
if(nargout > 2)
  data0 = data;
end

%---------------------------------------------------------------------------
%% Set up PDE term that implements the dynamics.

switch(dynamics_term)
  case 'convection'
    % The optimal feedback input for the double integrator can be determined 
    % analytically and thus the optimal dynamics can be implemented as a
    % purely convectional flow field.  Remember to reverse the dynamics
    % since we solve the PDE backward in time.
    schemeFunc = @termConvection;
    schemeData.grid = g;
    schemeData.derivFunc = derivFunc;
    schemeData.velocity = { -g.xs{2}; -(-input_bound * sign(g.xs{2})) };

  %case 'LaxFriedrichs'
    % A Lax-Friedrichs implementation would be more representative of what
    % would be used for more complicated dynamics, although it is likely to
    % be more dissipative and hence less accurate.
    
  otherwise
    error('Unknown dynamics_term: %s', dynamics_term);
end
    
%---------------------------------------------------------------------------
%% Restrict the Hamiltonian so that reachable set only grows.
% Whatever scheme handles the dynamics must already be completely set up.
if(restrict_update)
  innerFunc = schemeFunc;
  innerData = schemeData;
  clear schemeFunc schemeData;

  % Wrap the true Hamiltonian inside the term approximation restriction routine.
  schemeFunc = @termRestrictUpdate;
  schemeData.innerFunc = innerFunc;
  schemeData.innerData = innerData;
  schemeData.positive = 0;
end
  
%---------------------------------------------------------------------------
%% Initialize Display
f = figure; %#ok<NASGU>

% Set up subplot parameters if necessary.
if(useSubplots)
  rows = ceil(sqrt(plotSteps));
  cols = ceil(plotSteps / rows);
  plotNum = 1;
  subplot(rows, cols, plotNum);
end

h = visualizeLevelSet(g, data, displayType, level, [ 't = ' num2str(t0) ]);
set(h, 'LineStyle', '--', 'LineWidth', 1, 'LineColor', 'blue');
xlabel('position');  ylabel('velocity');

hold on;
if(g.dim > 1)
  axis(g.axis);
  daspect([ 1 1 1 ]);
end

%---------------------------------------------------------------------------
%% Loop until tMax (subject to a little roundoff).
tNow = t0;
startTime = cputime;
while(tMax - tNow > small * tMax)

  % Reshape data array into column vector for ode solver call.
  y0 = data(:);

  % How far to step?
  tSpan = [ tNow, min(tMax, tNow + tPlot) ];
  
  % Take a timestep.
  [ t y ] = feval(integratorFunc, schemeFunc, tSpan, y0,...
                  integratorOptions, schemeData);
  tNow = t(end);

  % Get back the correctly shaped data array
  data = reshape(y, g.shape);

  if(pauseAfterPlot)
    % Wait for last plot to be digested.
    pause;
  end

  % Delete last visualization if necessary.
  if(deleteLastPlot)
    delete(h);
  end

  % Move to next subplot if necessary.
  if(useSubplots)
    plotNum = plotNum + 1;
    subplot(rows, cols, plotNum);
  end

  % Create new visualization.
  h = visualizeLevelSet(g, data, displayType, level, [ 't = ' num2str(t) ]);
  set(h, 'LineStyle', ':', 'LineWidth', 1, 'LineColor', 'black');
  daspect([ 1 1 1 ]);
  
end

%---------------------------------------------------------------------------
%% Termination code.  
% Let the user know how long it all took.
endTime = cputime;
fprintf('Total execution time %g seconds\n', endTime - startTime);

% Make the final contour a little more visible.
set(h, 'LineStyle', '-', 'LineWidth', 2, 'LineColor', 'red');
title(sprintf('Double Integrator with implicit reachability at t = %4.2f with |u| \\leq %4.2f', tMax, input_bound));
